#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 13 07:50:55 2019

@author: fuaa
"""
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.cm as cm
import numpy as np
import getpass
import socket

_THRESHOLD = 10e-9

#########################################################################
def identificar_parcial():
    username = getpass.getuser()
    hostname = socket.gethostname()
    print("Usuario %s en %s." % (username, hostname))


##################################################################################
def printcolor( mensaje, color="k" ):
    """
    Función para imprimir en colores y con el formato de interés.
    """
    if   (color == "r"): mensajeColor = "\x1b[31m" + mensaje + "\x1b[0m"
    elif (color == "g"): mensajeColor = "\x1b[32m" + mensaje + "\x1b[0m"
    elif (color == "y"): mensajeColor = "\x1b[33m" + mensaje + "\x1b[0m"
    elif (color == "b"): mensajeColor = "\x1b[34m" + mensaje + "\x1b[0m"
    elif (color == "p"): mensajeColor = "\x1b[35m" + mensaje + "\x1b[0m"
    elif (color == "c"): mensajeColor = "\x1b[36m" + mensaje + "\x1b[0m"
    else: mensajeColor = mensaje
    mensaje_out = " " + mensajeColor 
    print ( mensaje_out )


##################################################################################
def fuaa_assert(condicion, 
                mensajeFalse = "El resultado no es válido.", 
                mensajeTrue = "Resultado validado." ):
    """
    Función similar al assert.
    """
    # Custom assert.
    if ( condicion ):
        printcolor( mensajeTrue, "g" )
    else:
        printcolor( mensajeFalse, "r" )


##################################################################################
def son_iguales(x1, x2, threshold = _THRESHOLD):
    """
    Evaluar si dos elementos son iguales o no, con una tolerancia dada (threshold).
    """
    dif = threshold + 1
    if x1.shape == x2.shape:
        if isinstance(x1, np.ndarray):
            dif = np.sqrt(np.sum( ( x1 - x2 )**2 )) / x1.size
        elif isinstance(x1, float):
            dif = np.abs( x1 - x2 )
        condicion = (dif < threshold)
    else:
        condicion = False
    return condicion

##################################################################################

def validar_parametros( parametros, min_params ):
    """
    Validar número de parámetros.
    """
    condicion = (len( parametros ) >= min_params)
    if not condicion:
        printcolor( "[validar_resultado] Insuficientes parámetros (\"%s\"), se necesitan %d, hay %d." % \
            (parametros[0], min_params, len(parametros)), "r" )
    return condicion

##################################################################################

def graficar_caracteristicas(Xb_train_features, yb_train, nombresClases):
    """
    Función para graficar las características de los datos.
    """

    plt.figure(figsize=(7, 7), dpi=120)
    plt.scatter(Xb_train_features[yb_train == 1, 1],
                Xb_train_features[yb_train == 1, 2],
                color='b',
                marker='o',
                label=nombresClases[0])
    
    plt.scatter(Xb_train_features[yb_train == -1, 1],
                Xb_train_features[yb_train == -1, 2],
                color='r',
                marker='x',
                label=nombresClases[1])

    plt.xlabel('Característica 1', fontsize=20)
    plt.ylabel('Característica 2', fontsize=20)
    plt.legend()
    plt.show()

##################################################################################

def graficar_dos_costos(costo1, costo2, label1="Costo 1", label2="Costo 2"):

    """
    Función para graficar dos costos.
    """
    
    # Obtener los límites máximo y mínimo de ambos costos
    min_costo = 0 # min(min(costo1), min(costo2))
    max_costo = max(max(costo1), max(costo2))

    # Crear una figura con dos subplots
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 5))

    # Subplot 1: Costo 1
    ax1.plot(costo1, label=label1, color='blue')
    ax1.set_title(label1)
    ax1.set_xlabel("Número de actualizaciones")
    ax1.set_ylabel("Valor del costo")
    ax1.grid()
    ax1.legend()

    # Subplot 2: Costo 2
    ax2.plot(costo2, label=label2, color='green')
    ax2.set_title(label2)
    ax2.set_xlabel("Número de actualizaciones")
    ax2.set_ylabel("Valor del costo")
    ax2.grid()
    ax2.legend()

    # Establecer los mismos límites en el eje Y para ambos subplots
    ax1.set_ylim(min_costo, max_costo)
    ax2.set_ylim(min_costo, max_costo)

    # Ajustar espacio entre subplots
    plt.tight_layout()

    # Mostrar la figura
    plt.show()

##################################################################################

def graficar_costos_entrenamiento(costo_train, costo_val):
    plt.figure(figsize=(10, 5))
    plt.plot(costo_train, label='Train', color='blue')
    plt.plot(costo_val, label='Val', color='red')
    plt.xlabel("Número de actualizaciones")
    plt.ylabel("Valor del costo")
    plt.grid()
    plt.legend()
    plt.show()

##################################################################################
def visualizar_modelo_v2(X, y, w_g, umbrales = [0.1, 0.25,0.5,0.75, 0.9], fig_size=(10, 10), title='Modelo lineal encontrado'):
    '''
    Entrada: 
        X: matriz de tamaño Nx3 que contiene el conjunto de puntos de entrenamiento expresados en
           en coordenadas homogeneas. La primera coordenada de cada punto es uno.
        y: etiquetas asignadas a los puntos
        w_g: parámetros del modelo lineal encontrados 
    '''
    plt.figure(figsize=fig_size)

    # Se grafican los puntos sorteados
    plt.scatter(X[y == -1, 1], X[y == -1, 2], s=40,
                color='r', marker='*', label='etiqueta -1')
    plt.scatter(X[y == 1, 1], X[y == 1, 2], s=40,
                color='b', marker='o', label='etiqueta 1')

    x1_min = X[:, 1].min()
    x1_max = X[:, 1].max()
    x1 = np.linspace(x1_min, x1_max)

    # Get unique colors for each umbral
    colors = cm.rainbow(np.linspace(0, 1, len(umbrales)))

    for umbral in umbrales:
        # Plot models
        if w_g[2] == 0:
            x2_g = -(w_g[0]/w_g[2])*np.ones(x1.shape)
        else:
            # Se grafica la superficie de decisión encontrada
            x2_g = - w_g[1]/w_g[2] * x1 - (w_g[0] - np.log((1-umbral)/umbral))/w_g[2]
        plt.plot(x1, x2_g, label='Funcion encontrada con umbral %0.2f' % umbral, color=colors[umbrales.index(umbral)])

    y1_min = X[:, 2].min()
    y1_max = X[:, 2].max()

    plt.legend()
    plt.axis('equal')
    plt.xlabel('x1')
    plt.ylabel('x2')
    plt.title(title)
    plt.xlim(x1_min, x1_max)
    plt.ylim(y1_min, y1_max)


##################################################################################
def validar_resultado(*args, **kwargs):
    """"
    Función para validar resultado a invocar desde el notebook.
    """
    # _ES_CORRECTO = False
    _DEBUG = False

    # Abrir el cartel del validación.
    print( "+-------------------------------------------------------------------------+" )
    print( "|               FuAA (1er. parcial 2023): validar resultado               |" )
    print( "+-------------------------------------------------------------------------+" )
    for key, value in kwargs.items():
        if key == "debug":
            _DEBUG = value
    if _DEBUG: 
        print('args:', args)
        print('kwargs:', kwargs)

    if ( len(args) == 0 ):
        print( "| Sin opciones para evaluar.                                              |" )
        print( "+-------------------------------------------------------------------------+" )
        return False
    
    ###########################################################
    # Ejercicio 3. Seccion 1 - Parte A 
    ###########################################################
    elif (args[0] == "sec1_a"):
        if validar_parametros( args, 4 ):

            Xb_train = args[1]
            yb_train = args[2]
            Xb_val = args[3]
            yb_val = args[4]

            condicion1 = Xb_train.shape[0] == 1000

            fuaa_assert(   condicion1, \
            mensajeFalse = "El conjunto de Entrenamiento no tiene el tamaño correcto", \
            mensajeTrue = "El conjunto de Entremaniento tiene el tamaño correcto.")
           
           
            condicion2 = len(Xb_train)==len(yb_train)

            fuaa_assert(   condicion2, \
                        mensajeFalse = "El número de etiquetas de entrenamiento debe coincidir con el número de dígitos.", \
                        mensajeTrue = "Misma cantidad de datos que etiquetas en Entrenamiento.")
            
            condicion3 = len(Xb_val)==len(yb_val)

            fuaa_assert(   condicion3, \
                        mensajeFalse = "El número de etiquetas de validación debe coincidir con el número de dígitos.", \
                        mensajeTrue = "Misma cantidad de datos que etiquetas en Validación.")

    ###########################################################
    # Ejercicio 3. Seccion 2 - Parte A - Feature 1
    ###########################################################         
    elif (args[0] == "sec2_a_feature1"):
        if 'funcion' in kwargs:
            feature1 = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return
        
        # Set seed for reproducibility
        np.random.seed(42)

        # Generate random image
        image1 = np.random.randint(0, 16, 256)

        resultado1 = feature1(image1)

        # Set seed for reproducibility
        np.random.seed(43)

        # Generate random image
        image2 = np.random.randint(0, 16, 256)

        resultado2 = feature1(image2)

        x1 = np.array(-0.109375)
        x2 = np.array(0.8125)

        condicion1 = son_iguales(resultado1, x1) and son_iguales(resultado2, x2)

        fuaa_assert(   condicion1, \
                    mensajeFalse = "Revisar la implementación de la Feature 1", \
                    mensajeTrue = "Implementación de la Feature 1: resultado validado.")
        
    
    ###########################################################
    # Ejercicio 3. Seccion 2 - Parte A - Feature 2
    ###########################################################         
    elif (args[0] == "sec2_a_feature2"):
        if 'funcion' in kwargs:
            feature2 = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        # Set seed for reproducibility
        np.random.seed(42)

        # Generate random image
        image1 = np.random.randint(0, 16, 256)

        resultado1 = feature2(image1)

        # Set seed for reproducibility
        np.random.seed(43)

        # Generate random image
        image2 = np.random.randint(0, 16, 256)

        resultado2 = feature2(image2)

        x1 = np.array(-0.03125)
        x2 = np.array(-0.703125)

        condicion1 = son_iguales(resultado1, x1) and son_iguales(resultado2, x2)

        fuaa_assert(   condicion1, \
                    mensajeFalse = "Revisar la implementación de la Feature 2", \
                    mensajeTrue = "Implementación de la Feature 2: resultado validado.")


    ###########################################################
    # Ejercicio 2. Costo regresión logística
    ###########################################################
    
    elif (args[0] == "calcular_costo"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(33)
        X5 = np.random.randn(3,4)
        X5[:,0]=1
        y5 = np.array([1,-1,1])
        w5 = np.random.randn(4)

        c5 = test_algoritmo(X5, y5, w5 )
        # print(c5)
        condicion1 =  np.abs(c5 - 0.13048058961396589) < _THRESHOLD
        fuaa_assert(   condicion1, \
                    mensajeFalse = "Revisar la implementación del costo de la regresión logística", \
                    mensajeTrue = "Implementación del costo de la regresión logística: resultado validado.")

    
    ###########################################################
    # Ejercicio 2. Gradiente regresión logística
    ###########################################################
    elif (args[0] == "calcular_gradiente"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(33)
        X5 = np.random.randn(3,4)
        X5[:,0]=1
        y5 = np.array([1,-1,1])
        w5 = np.random.randn(4)

        g5 = test_algoritmo(X5,y5,w5) 
        condicion1  = g5.shape == w5.shape
        fuaa_assert(condicion1, \
                    mensajeFalse = "Revisar las dimensiones del gradiente", \
                    mensajeTrue = "Dimensiones del gradiente son correctas.")
        condicion2 = son_iguales(g5, np.array([-0.10645973,  0.10632848,  0.12291721,  0.07835299]))
        fuaa_assert(   condicion2, \
                    mensajeFalse = "Revisar la implementación del gradiente de la regresión logística", \
                    mensajeTrue = "Implementación del gradiente de la regresión logística: resultado validado.")

   
    
    else:
        printcolor( "Ninguna opción revisada." ) 

    # Cerrar el cartel.
    print( "+-------------------------------------------------------------------------+" )
            
# condicion = False
# mensaje = "Este ese el texto a mostrar en caso de condición falsa."
# validar_resultado( "test", condicion, mensaje )


